"""
BTrees are commonly used in the Zope world.  This modules exposes the
contents of BTrees nicely, abstracting away the implementation details.

In the DB, every BTree can be represented by more than one persistent object,
every one of those versioned separately.  This is part of what makes BTrees
efficient.

The format of the picked BTree state is nicely documented in ZODB's source
code, specifically, BTreeTemplate.c and BucketTemplate.c.
"""

from BTrees.OOBTree import OOBTree, OOBucket
from zope.component import adapter, getMultiAdapter
from zope.interface import implementer


# be compatible with Zope 3.4, but prefer the modern package structure
try:
    from zope.container.folder import Folder
except ImportError:
    from zope.app.folder import Folder  # BBB
try:
    from zope.container.btree import BTreeContainer
except ImportError:
    from zope.app.container.btree import BTreeContainer # BBB

from zodbbrowser.history import ZodbObjectHistory
from zodbbrowser.interfaces import IObjectHistory, IStateInterpreter
from zodbbrowser.state import GenericState


@adapter(OOBTree)
@implementer(IObjectHistory)
class OOBTreeHistory(ZodbObjectHistory):

    def _load(self):
        # find all objects (tree and buckets) that have ever participated in
        # this OOBTree
        queue = [self._obj]
        seen = set(self._oid)
        history_of = {}
        while queue:
            obj = queue.pop(0)
            history = history_of[obj._p_oid] = ZodbObjectHistory(obj)
            for d in history:
                state = history.loadState(d['tid'])
                if state and len(state) > 1:
                    bucket = state[1]
                    if bucket._p_oid not in seen:
                        queue.append(bucket)
                        seen.add(bucket._p_oid)
        # merge the histories of all objects
        by_tid = {}
        for h in history_of.values():
            for d in h:
                by_tid.setdefault(d['tid'], d)
        self._history = sorted(by_tid.values(),
                               key=lambda d: d['tid'], reverse=True)
        self._index_by_tid()

    def _lastRealChange(self, tid=None):
        return ZodbObjectHistory(self._obj).lastChange(tid)

    def loadStatePickle(self, tid=None):
        # lastChange would return the tid that modified self._obj or any
        # of its subobjects, thanks to the history merging done by _load.
        # We need the real last change value.
        # XXX: this is used to show the pickled size of an object.  It
        # will be misleading for BTrees if we show just the size for the
        # main BTree object while we're hiding all the individual buckets.
        return self._connection._storage.loadSerial(self._obj._p_oid,
                                                    self._lastRealChange(tid))

    def loadState(self, tid=None):
        # lastChange would return the tid that modified self._obj or any
        # of its subobjects, thanks to the history merging done by _load.
        # We need the real last change value.
        return self._connection.oldstate(self._obj, self._lastRealChange(tid))

    def rollback(self, tid):
        state = self.loadState(tid)
        if state != self.loadState():
            self._obj.__setstate__(state)
            self._obj._p_changed = True

        while state and len(state) > 1:
            bucket = state[1]
            bucket_history = IObjectHistory(bucket)
            state = bucket_history.loadState(tid)
            if state != bucket_history.loadState():
                bucket.__setstate__(state)
                bucket._p_changed = True


@adapter(OOBTree, tuple, None)
@implementer(IStateInterpreter)
class OOBTreeState(object):
    """Non-empty OOBTrees have a complicated tuple structure."""

    def __init__(self, type, state, tid):
        self._items = []
        # A BTree has two kinds of nodes: internal nodes are OOBTree objects,
        # leaf nodes are OOBucket objects.  Internal nodes contain
        # either OOBTrees or OOBuckets (or nothing, if it's an empty node).
        # Buckets contain actual keys and values.  In addition, each internal
        # node has a firstbucket reference that points to the leftmost
        # grandchild, and each bucket has a nextbucket reference.  We
        # can start at firstbucket and follow nextbucket pointers and we'll
        # get all the items in increasing order by key.
        #
        # The pickled state of an OOBTree can be:
        # - `None` for an empty tree
        # - `((first_bucket_state, ), )` for a tree with one bucket
        # - `((child0, key1, child1, ...), firstbucket)` for the general case
        #
        # The pickled state of an OOBucket is
        # - `((key0, value0, key1, value1, ...),)` for the last bucket
        # - `((key0, value0, key1, value1, ...), next_bucket)` for the general case
        #
        # This is documented in comments in BTreeTemplate.c and BucketTemplate.c.
        if not state:
            state = ()
        else:
            assert isinstance(state, tuple)
        if len(state) == 1:
            firstbucket_state = state[0][0]
        elif len(state) == 2:
            bucket = state[1]
            firstbucket_state = IObjectHistory(bucket).loadState(tid)
        else:
            assert len(state) == 0
            firstbucket_state = ((),)

        state = firstbucket_state
        while True:
            assert isinstance(state, tuple)
            assert 1 <= len(state) <= 2
            items = state[0]
            assert isinstance(items, tuple)
            assert len(items) % 2 == 0
            it = iter(items)
            self._items.extend(zip(it, it))

            if len(state) == 1:
                break

            bucket = state[1]
            state = IObjectHistory(bucket).loadState(tid)

    def getError(self):
        return None

    def getName(self):
        return None

    def getParent(self):
        return None

    def listAttributes(self):
        return None

    def listItems(self):
        return self._items

    def asDict(self):
        return dict(self._items)


@adapter(OOBTree, type(None), None)
@implementer(IStateInterpreter)
class EmptyOOBTreeState(OOBTreeState):
    """Empty OOBTrees pickle to None."""


@adapter(Folder, dict, None)
class FolderState(GenericState):
    """Convenient access to a Folder's items"""

    def listItems(self):
        data = self.state.get('data')
        if not data:
            return []
        # data will be an OOBTree
        loadedstate = IObjectHistory(data).loadState(self.tid)
        return getMultiAdapter((data, loadedstate, self.tid),
                               IStateInterpreter).listItems()


@adapter(BTreeContainer, dict, None)
class BTreeContainerState(GenericState):
    """Convenient access to a BTreeContainer's items"""

    def listItems(self):
        # This is not a typo; BTreeContainer really uses
        # _SampleContainer__data, for BBB
        data = self.state.get('_SampleContainer__data')
        if not data:
            return []
        # data will be an OOBTree
        loadedstate = IObjectHistory(data).loadState(self.tid)
        return getMultiAdapter((data, loadedstate, self.tid),
                               IStateInterpreter).listItems()


@adapter(OOBucket, tuple, None)
@implementer(IStateInterpreter)
class OOBucketState(GenericState):
    """A single OOBTree bucket, should you wish to look at the internals

    Here's the state description direct from BTrees/BucketTemplate.c::

     * For a set bucket (self->values is NULL), a one-tuple or two-tuple.  The
     * first element is a tuple of keys, of length self->len.  The second element
     * is the next bucket, present if and only if next is non-NULL:
     *
     *     (
     *          (keys[0], keys[1], ..., keys[len-1]),
     *          <self->next iff non-NULL>
     *     )
     *
     * For a mapping bucket (self->values is not NULL), a one-tuple or two-tuple.
     * The first element is a tuple interleaving keys and values, of length
     * 2 * self->len.  The second element is the next bucket, present iff next is
     * non-NULL:
     *
     *     (
     *          (keys[0], values[0], keys[1], values[1], ...,
     *                               keys[len-1], values[len-1]),
     *          <self->next iff non-NULL>
     *     )

    OOBucket is a mapping bucket; OOSet is a set bucket.
    """

    def getError(self):
        return None

    def getName(self):
        return None

    def getParent(self):
        return None

    def listAttributes(self):
        return [('_next', self.state[1] if len(self.state) > 1 else None)]

    def listItems(self):
        return zip(self.state[0][::2], self.state[0][1::2])

    def asDict(self):
        return dict(self.listAttributes(), _items=dict(self.listItems()))




from persistent import Persistent
from ZODB.interfaces import IConnection, IStorageIteration
from ZODB.utils import tid_repr
from zope.component import adapter
from zope.interface import implementer
from zope.proxy import removeAllProxies

from zodbbrowser import cache
from zodbbrowser.interfaces import IDatabaseHistory, IObjectHistory


try:
    from ZODB.mvccadapter import MVCCAdapterInstance
except ImportError:  # pragma: no-cover
    class MVCCAdapterInstance(object):
        """Placeholder so we can register an adapter that will not be used."""


@adapter(Persistent)
@implementer(IObjectHistory)
class ZodbObjectHistory(object):

    def __init__(self, obj):
        self._obj = removeAllProxies(obj)
        self._connection = self._obj._p_jar
        self._storage = self._connection._storage
        self._oid = self._obj._p_oid
        self._history = None
        self._by_tid = {}

    def __len__(self):
        if self._history is None:
            self._load()
        return len(self._history)

    def _load(self):
        """Load history of changes made to a Persistent object.

        Returns a list of dictionaries, from latest revision to the oldest.
        The dicts have various interesting pieces of data, such as:

            tid -- transaction ID (a byte string, usually 8 bytes)
            time -- transaction timestamp (number of seconds since the Unix epoch)
            user_name -- name of the user responsible for the change
            description -- short description (often a URL)

        See the 'history' method of ZODB.interfaces.IStorage.
        """
        size = 999999999999 # "all of it"; ought to be sufficient
        self._history = self._storage.history(self._oid, size=size)
        self._index_by_tid()

    def _index_by_tid(self):
        for record in self._history:
            self._by_tid[record['tid']] = record

    def __getitem__(self, item):
        if self._history is None:
            self._load()
        d = dict(self._history[item])
        if isinstance(d['user_name'], bytes):
            d['user_name'] = d['user_name'].decode('UTF-8', 'replace')
        if isinstance(d['description'], bytes):
            d['description'] = d['description'].decode('UTF-8', 'replace')
        return d

    def lastChange(self, tid=None):
        if self._history is None:
            self._load()
        if tid in self._by_tid:
            # optimization
            return tid
        # sadly ZODB has no API for get revision at or before tid, so
        # we have to find the exact tid
        for record in self._history:
            # we assume records are ordered by tid, newest to oldest
            if tid is None or record['tid'] <= tid:
                return record['tid']
        raise KeyError('%r did not exist in or before transaction %r' %
                       (self._obj, tid_repr(tid)))

    def loadStatePickle(self, tid=None):
        return self._connection._storage.loadSerial(self._obj._p_oid,
                                                    self.lastChange(tid))

    def loadState(self, tid=None):
        return self._connection.oldstate(self._obj, self.lastChange(tid))

    def rollback(self, tid):
        state = self.loadState(tid)
        if state != self.loadState():
            self._obj.__setstate__(state)
            self._obj._p_changed = True


def getObjectHistory(obj):
    assert isinstance(obj, Persistent)
    history = IObjectHistory(obj, None)
    if history is None:
        # See LP: #1185175
        history = ZodbObjectHistory(obj)
    return history


@adapter(IConnection)
@implementer(IDatabaseHistory)
class ZodbHistory(object):

    def __init__(self, connection):
        self._connection = connection
        self._storage = IStorageIteration(connection._storage)
        self._tids = cache.getStorageTids(self._storage)
        self._iterators = []

    @property
    def tids(self):
        return tuple(self._tids) # readonlify

    def __len__(self):
        return len(self._tids)

    def _addcleanup(self, it):
        self._iterators.append(it)
        return it

    def cleanup(self):
        for it in self._iterators:
            if hasattr(it, 'close'):
                it.close()
        self._iterators = []

    def __iter__(self):
        return self._addcleanup(self._storage.iterator())

    def __getitem__(self, index):
        if isinstance(index, slice):
            assert index.step is None or index.step == 1
            tids = self._tids[index]
            if not tids:
                return []
            return self._addcleanup(self._storage.iterator(tids[0], tids[-1]))
        else:
            tid = self._tids[index]
            return next(self._addcleanup(self._storage.iterator(tid, tid)))


@adapter(MVCCAdapterInstance)
@implementer(IStorageIteration)
def getIterableStorage(storage):
    return storage._storage



import transaction
from persistent import Persistent
from persistent.dict import PersistentDict
from zope.component import provideAdapter
from zope.interface.verify import verifyObject

from zodbbrowser.history import (
    ZodbHistory,
    ZodbObjectHistory,
    getIterableStorage,
)
from zodbbrowser.interfaces import IDatabaseHistory, IObjectHistory
from zodbbrowser.tests.realdb import RealDatabaseTest


class PersistentObject(Persistent):
    pass # we need a subclass so we get a __dict__


class TestFileStorage(RealDatabaseTest):

    def setUp(self):
        RealDatabaseTest.setUp(self)
        self.obj = self.conn.root()['obj'] = PersistentDict()
        transaction.commit()

    def test_no_history(self):
        obj = self.obj
        history = ZodbObjectHistory(obj)
        verifyObject(IObjectHistory, history)
        self.assertEqual(len(history), 1)
        self.assertTrue('tid' in history[0])
        self.assertTrue('time' in history[0])
        self.assertTrue('user_name' in history[0])
        self.assertTrue('description' in history[0])
        self.assertNotIsInstance(history[0]['user_name'], bytes)
        self.assertNotIsInstance(history[0]['description'], bytes)

    def test_some_history(self):
        obj = self.obj
        for n in range(10):
            obj[n] = n
            transaction.commit()
        history = ZodbObjectHistory(obj)
        self.assertEqual(len(history), 11)


class WorkloadMixin(object):

    def commitSomeStuff(self):
        root = self.conn.root()
        self.adam = root['adam'] = PersistentObject()
        transaction.commit()
        self.eve = root['eve'] = PersistentObject()
        transaction.commit()
        self.adam.laptop = 'ThinkPad T23'
        transaction.commit()
        self.eve.laptop = 'MacBook'
        transaction.commit()
        self.adam.laptop = 'ThinkPad T42'
        transaction.commit()
        self.adam.laptop = 'ThinkPad T61'
        transaction.commit()


class TestLoadState(WorkloadMixin, RealDatabaseTest):

    def setUp(self):
        RealDatabaseTest.setUp(self)
        self.commitSomeStuff()

    def test_latest_state(self):
        state = ZodbObjectHistory(self.adam).loadState()
        self.assertEqual(state, dict(laptop='ThinkPad T61'))

    def test_exact_state(self):
        tid = ZodbObjectHistory(self.adam)[1]['tid']
        state = ZodbObjectHistory(self.adam).loadState(tid)
        self.assertEqual(state, dict(laptop='ThinkPad T42'))

    def test_earlier_state(self):
        tid = ZodbObjectHistory(self.eve)[0]['tid']
        state = ZodbObjectHistory(self.adam).loadState(tid)
        self.assertEqual(state, dict(laptop='ThinkPad T23'))

    def test_error_handling(self):
        tid = ZodbObjectHistory(self.adam)[-1]['tid']
        history = ZodbObjectHistory(self.eve)
        try:
            history.loadState(tid)
        except KeyError as e:
            self.assertTrue("did not exist in or before" in str(e))
        else:
            self.fail("did not raise")

    def test_rollback_does_nothing(self):
        history = ZodbObjectHistory(self.adam)
        history.rollback(history.lastChange())
        self.assertEqual(self.adam.laptop, 'ThinkPad T61')
        self.assertFalse(self.adam._p_changed)

    def test_rollback(self):
        history = ZodbObjectHistory(self.adam)
        history.rollback(history[-2]['tid'])
        self.assertEqual(self.adam.laptop, 'ThinkPad T23')
        self.assertTrue(self.adam._p_changed)


class TestZodbHistory(WorkloadMixin, RealDatabaseTest):

    def setUp(self):
        super(TestZodbHistory, self).setUp()
        provideAdapter(getIterableStorage)

    def test_no_history(self):
        history = ZodbHistory(self.conn)
        self.addCleanup(history.cleanup)
        verifyObject(IDatabaseHistory, history)
        self.assertEqual(len(history), 1)
        self.assertEqual(len(history.tids), 1)
        self.assertEqual(len(list(history)), 1)
        self.assertEqual([tr.tid for tr in history],
                         [tr.tid for tr in history[-5:]])
        self.assertEqual(history[10:], [])
        self.assertEqual(history[0].tid, history.tids[0])

    def test_some_history(self):
        self.commitSomeStuff()
        history = ZodbHistory(self.conn)
        self.addCleanup(history.cleanup)
        verifyObject(IDatabaseHistory, history)
        self.assertEqual(len(history), 7)
        self.assertEqual(len(list(history)), 7)
        self.assertEqual([tr.tid for tr in history][-5:],
                         [tr.tid for tr in history[-5:]])

